/*
 * 	cn_queue.c
 * 
 * 2004 Copyright (c) Evgeniy Polyakov <johnpol@2ka.mipt.ru>
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/skbuff.h>
#include <linux/suspend.h>
#include <linux/delay.h>

#include "connector.h"

static void cn_queue_wrapper(void *data)
{
	struct cn_callback_entry *cbq = (struct cn_callback_entry *)data;

	atomic_inc(&cbq->cb->refcnt);
	cbq->cb->callback(cbq->cb->priv);
	atomic_dec(&cbq->cb->refcnt);

	cbq->destruct_data(cbq->ddata);
}

static struct cn_callback_entry *cn_queue_alloc_callback_entry(struct
							       cn_callback *cb)
{
	struct cn_callback_entry *cbq;

	cbq = kmalloc(sizeof(*cbq), GFP_KERNEL);
	if (!cbq) {
		printk(KERN_ERR "Failed to create new callback queue.\n");
		return NULL;
	}

	memset(cbq, 0, sizeof(*cbq));

	cbq->cb = cb;

	INIT_WORK(&cbq->work, &cn_queue_wrapper, cbq);

	return cbq;
}

static void cn_queue_free_callback(struct cn_callback_entry *cbq)
{
	cancel_delayed_work(&cbq->work);

	while (atomic_read(&cbq->cb->refcnt)) {
		printk(KERN_INFO "Waiting for %s to become free: refcnt=%d.\n",
		       cbq->pdev->name, atomic_read(&cbq->cb->refcnt));

		msleep_interruptible(1000);

		if (current->flags & PF_FREEZE)
			refrigerator(PF_FREEZE);

		if (signal_pending(current))
			flush_signals(current);
	}

	kfree(cbq);
}

int cn_cb_equal(struct cb_id *i1, struct cb_id *i2)
{
#if 0
	printk(KERN_INFO "%s: comparing %04x.%04x and %04x.%04x\n",
			__func__,
			i1->idx, i1->val,
			i2->idx, i2->val);
#endif
	return ((i1->idx == i2->idx) && (i1->val == i2->val));
}

int cn_queue_add_callback(struct cn_queue_dev *dev, struct cn_callback *cb)
{
	struct cn_callback_entry *cbq, *n, *__cbq;
	int found = 0;

	cbq = cn_queue_alloc_callback_entry(cb);
	if (!cbq)
		return -ENOMEM;

	atomic_inc(&dev->refcnt);
	cbq->pdev = dev;

	spin_lock_bh(&dev->queue_lock);
	list_for_each_entry_safe(__cbq, n, &dev->queue_list, callback_entry) {
		if (cn_cb_equal(&__cbq->cb->id, &cb->id)) {
			found = 1;
			break;
		}
	}
	if (!found) {
		atomic_set(&cbq->cb->refcnt, 1);
		list_add_tail(&cbq->callback_entry, &dev->queue_list);
	}
	spin_unlock_bh(&dev->queue_lock);

	if (found) {
		atomic_dec(&dev->refcnt);
		atomic_set(&cbq->cb->refcnt, 0);
		cn_queue_free_callback(cbq);
		return -EINVAL;
	}

	cbq->nls = dev->nls;
	cbq->seq = 0;
	cbq->group = cbq->cb->id.idx;

	return 0;
}

void cn_queue_del_callback(struct cn_queue_dev *dev, struct cn_callback *cb)
{
	struct cn_callback_entry *cbq = NULL, *n;
	int found = 0;

	spin_lock_bh(&dev->queue_lock);
	list_for_each_entry_safe(cbq, n, &dev->queue_list, callback_entry) {
		if (cn_cb_equal(&cbq->cb->id, &cb->id)) {
			list_del(&cbq->callback_entry);
			found = 1;
			break;
		}
	}
	spin_unlock_bh(&dev->queue_lock);

	if (found) {
		atomic_dec(&cbq->cb->refcnt);
		cn_queue_free_callback(cbq);
		atomic_dec(&dev->refcnt);
	}
}

struct cn_queue_dev *cn_queue_alloc_dev(char *name, struct sock *nls)
{
	struct cn_queue_dev *dev;

	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev) {
		printk(KERN_ERR "%s: Failed to allocte new struct cn_queue_dev.\n",
		       name);
		return NULL;
	}

	memset(dev, 0, sizeof(*dev));

	snprintf(dev->name, sizeof(dev->name), "%s", name);

	atomic_set(&dev->refcnt, 0);
	INIT_LIST_HEAD(&dev->queue_list);
	spin_lock_init(&dev->queue_lock);

	dev->nls = nls;
	dev->netlink_groups = 0;

	dev->cn_queue = create_workqueue(dev->name);
	if (!dev->cn_queue) {
		printk(KERN_ERR "Failed to create %s queue.\n", dev->name);
		kfree(dev);
		return NULL;
	}

	return dev;
}

void cn_queue_free_dev(struct cn_queue_dev *dev)
{
	struct cn_callback_entry *cbq, *n;

	flush_workqueue(dev->cn_queue);
	destroy_workqueue(dev->cn_queue);

	spin_lock_bh(&dev->queue_lock);
	list_for_each_entry_safe(cbq, n, &dev->queue_list, callback_entry) {
		list_del(&cbq->callback_entry);
		atomic_dec(&cbq->cb->refcnt);
	}
	spin_unlock_bh(&dev->queue_lock);

	while (atomic_read(&dev->refcnt)) {
		printk(KERN_INFO "Waiting for %s to become free: refcnt=%d.\n",
		       dev->name, atomic_read(&dev->refcnt));

		msleep_interruptible(1000);

		if (current->flags & PF_FREEZE)
			refrigerator(PF_FREEZE);

		if (signal_pending(current))
			flush_signals(current);
	}

	memset(dev, 0, sizeof(*dev));
	kfree(dev);
	dev = NULL;
}
