/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * Copyright (C) 2004 Mike Christie
 * Copyright (C) 2004 IBM Corporation
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-session.h,v 1.60 2005/01/11 04:25:19 mikenc Exp $
 *
 * define the iSCSI session structure needed by the login library
 */
#ifndef ISCSI_SESSION_H_
#define ISCSI_SESSION_H_

#include <linux/crypto.h>
#include <linux/socket.h>
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>

#include "iscsi-auth-client.h"
#include "iscsi-portal.h"

struct iscsi_session_ioctl;
struct iscsi_task;

/* used for replying to NOPs */
struct iscsi_nop_info {
	struct list_head reply_list;
	u32 ttt;
	unsigned char lun[8];
};

#define ISCSI_RXCTRL_SIZE PAGE_SIZE

struct iscsi_session {
	struct Scsi_Host *shost;
	struct list_head list;
	struct socket *socket;
	struct sockaddr addr;
	int tcp_window_size;
	struct semaphore config_mutex;
	u32 config_number;
	char *username;
	unsigned char *password;
	int password_length;
	char *username_in;
	unsigned char *password_in;
	int password_length_in;
	unsigned char *initiator_name;
	unsigned char *initiator_alias;
	unsigned char *target_name;
	unsigned char *target_alias;
	struct crypto_tfm *md5_tfm;
	int bidirectional_auth;
	struct iscsi_acl *auth_client_block;
	struct auth_str_block *auth_recv_string_block;
	struct auth_str_block *auth_send_string_block;
	struct auth_large_binary *auth_recv_binary_block;
	struct auth_large_binary *auth_send_binary_block;
	int num_auth_buffers;
	struct auth_buffer_desc auth_buffers[5];
	spinlock_t portal_lock;
	struct iscsi_portal_info portal;
	int ever_established;
	int session_alive;
	int commands_queued;
	struct timer_list session_timer;
	u8 isid[6];
	u16 tsih;
	unsigned int cmd_sn;
	u32 exp_cmd_sn;
	u32 max_cmd_sn;
	int immediate_data;
	int initial_r2t;
	int max_recv_data_segment_len;	/* the value we declare */
	int max_xmit_data_segment_len;	/* the value declared by the target */
	int first_burst_len;
	int max_burst_len;
	int data_pdu_in_order;
	int data_seq_in_order;
	int def_time2wait;
	int def_time2retain;
	int header_digest;
	int data_digest;
	int type;
	int current_stage;
	int next_stage;
	int partial_response;
	int portal_group_tag;
	u32 next_itt;
	unsigned long last_rx;
	unsigned long last_ping;
	unsigned long last_window_check;
	unsigned long last_kill;
	unsigned long login_phase_timer;
	unsigned long window_full;
	unsigned long window_closed;
	int vendor_specific_keys;
	unsigned int irrelevant_keys_bitmap;
	int send_async_text;
	int login_timeout;
	int auth_timeout;
	int active_timeout;
	int idle_timeout;
	int ping_timeout;
	int abort_timeout;
	int reset_timeout;
	int replacement_timeout;
	/*
	 * the queue of SCSI commands that have been sent on this session, and
	 * for which we're waiting for a reply.
	 */
	spinlock_t task_lock;
	struct iscsi_task *preallocated_task;
	struct list_head pending_queue;
	struct list_head active_queue;
	struct list_head done_queue;
	struct list_head tx_task_head;
	atomic_t num_active_tasks;
	wait_queue_head_t queue_empty;
	struct iscsi_nop_info nop_reply;
	struct list_head nop_reply_list;
	u32 last_mgmt_itt; /* itt of the last mgmt task we sent */
	struct iscsi_task *mgmt_task; /* preallocated task for TMFs */
	struct completion *mgmt_task_complete;
	wait_queue_head_t tx_wait_q;
	wait_queue_head_t login_wait_q;
	struct semaphore tx_blocked;
	unsigned long control_bits;
	struct task_struct *rx_task;
	struct task_struct *tx_task;
	unsigned long session_drop_time;
	unsigned long session_established_time;
	/* the following fields are per-connection, not per session, and will
	 * need to move if we decide to support multiple connections per
	 * session.
	 */
	unsigned long logout_deadline;
	unsigned long logout_response_deadline;
	u32 logout_itt;
	long time2wait;
	unsigned int exp_stat_sn;
	unsigned char rx_buffer[ISCSI_RXCTRL_SIZE];
	struct crypto_tfm *rx_tfm;
	struct crypto_tfm *tx_tfm;
};

/* session control bits */
enum {
	/*
	 * the tx bits match the tx_request array in
	 * iscsi-initiator.c, so if you modify this don't forget
	 */
	TX_PING,		/* NopOut, reply requested */
	TX_TMF,
	TX_SCSI_COMMAND,
	TX_NOP_REPLY,		/* reply to a Nop-in from the target */
	TX_DATA,
	TX_LOGOUT,
	TX_WAKE,

	SESSION_CREATED,
	SESSION_RELEASING,
	SESSION_REPLACEMENT_TIMEDOUT,
	SESSION_ESTABLISHED,
	SESSION_LOGOUT_REQUESTED,
	SESSION_WINDOW_CLOSED,
	SESSION_TERMINATING,
	SESSION_TERMINATED,
};

extern void iscsi_wake_tx_thread(int control_bit,
				 struct iscsi_session *session);
extern void iscsi_request_logout(struct iscsi_session *session, int logout,
				 int logout_response);
extern void iscsi_drop_session(struct iscsi_session *session);
extern int iscsi_update_session(struct iscsi_session *session,
				struct iscsi_session_ioctl *ioctld);
extern int iscsi_create_session(struct iscsi_session *session,
				struct iscsi_session_ioctl *ioctld);
extern void iscsi_destroy_session(struct iscsi_session *session);
extern struct iscsi_session *iscsi_find_session(const char *target_name,
						u8 isid[6], int tpgt);
extern int iscsi_update_address(struct iscsi_session *session, char *address);
extern int iscsi_wait_for_session(struct iscsi_session *session,
				  int ignore_timeout);
extern struct list_head iscsi_sessions;

#endif
