/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * Copyright (C) 2004 Mike Christie
 * Copyright (C) 2004 IBM Corporation
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-session.c,v 1.77 2005/01/11 18:50:10 mikenc Exp $
 *
 * This File implements the funtions related to establishing and
 * managing the session.
 */
#include <linux/blkdev.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/inet.h>
#include <linux/interrupt.h>
#include <scsi/scsi_device.h>

#include "iscsi-session.h"
#include "iscsi-ioctl.h"
#include "iscsi-task.h"
#include "iscsi-login.h"
#include "iscsi-sfnet.h"

#define DEF_SESSION_TIMEOUT (jiffies + (5 * HZ))

/*
 * list of initialized iscsi sessions - this should be replaced
 * with a driver model equivalent if possible.
 */
LIST_HEAD(iscsi_sessions);
static DECLARE_MUTEX(iscsi_session_sem);

static void
signal_iscsi_threads(struct iscsi_session *session)
{
	if (session->tx_task)
		kill_proc(session->tx_task->pid, SIGHUP, 1);
	if (session->rx_task)
		kill_proc(session->rx_task->pid, SIGHUP, 1);
}

/* drop an iscsi session */
void
iscsi_drop_session(struct iscsi_session *session)
{
	if (!test_and_clear_bit(SESSION_ESTABLISHED, &session->control_bits))
		return;

	/* so we know whether to abort the connection */
	session->session_drop_time = jiffies ? jiffies : 1;
	signal_iscsi_threads(session);
	session->session_alive = 0;
}

/* caller must hold session->task_lock */
void
iscsi_request_logout(struct iscsi_session *session, int logout,
		     int logout_response)
{
	if (atomic_read(&session->num_active_tasks) == 0) {
		session->logout_response_deadline =
		    jiffies + (logout_response * HZ);
		if (session->logout_response_deadline == 0)
			session->logout_response_deadline = 1;
		set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
		iscsi_wake_tx_thread(TX_LOGOUT, session);
	} else {
		session->logout_deadline = jiffies + (logout * HZ);
		if (session->logout_deadline == 0)
			session->logout_deadline = 1;
		session->logout_response_deadline =
		    session->logout_deadline + (logout_response * HZ);
		if (session->logout_response_deadline == 0)
			session->logout_response_deadline = 1;
		set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
	}
}

static void
free_session(struct iscsi_session *session)
{
	if (session->preallocated_task)
		kmem_cache_free(iscsi_task_cache, session->preallocated_task);

	if (session->mgmt_task)
		kmem_cache_free(iscsi_task_cache, session->mgmt_task);

	/* free crypto structures */
	if (session->rx_tfm)
		crypto_free_tfm(session->rx_tfm);
	if (session->tx_tfm)
		crypto_free_tfm(session->tx_tfm);
	if (session->md5_tfm)
		crypto_free_tfm(session->md5_tfm);
	/* free the auth structures */
	kfree(session->auth_client_block);
	kfree(session->auth_recv_string_block);
	kfree(session->auth_send_string_block);
	kfree(session->auth_recv_binary_block);
	kfree(session->auth_send_binary_block);
	kfree(session->username);
	kfree(session->password);
	kfree(session->username_in);
	kfree(session->password_in);
	kfree(session->initiator_name);
	kfree(session->initiator_alias);
	kfree(session->target_name);
	kfree(session->target_alias);
}

static void
assign_auth_buffers(struct iscsi_session *session)
{
	session->auth_buffers[0].address = session->auth_client_block;
	session->auth_buffers[0].length = sizeof(*session->auth_client_block);
	session->auth_buffers[1].address = session->auth_recv_string_block;
	session->auth_buffers[1].length =
		sizeof(*session->auth_recv_string_block);
	session->auth_buffers[2].address = session->auth_send_string_block;
	session->auth_buffers[2].length =
		sizeof(*session->auth_send_string_block);
	session->auth_buffers[3].address = session->auth_recv_binary_block;
	session->auth_buffers[3].length =
		sizeof(*session->auth_recv_binary_block);
	session->auth_buffers[4].address = session->auth_send_binary_block;
	session->auth_buffers[4].length =
		sizeof(*session->auth_send_binary_block);
	session->num_auth_buffers = 5;
}

static int
allocate_auth_buffers(struct iscsi_session *session)
{
	if (!(session->bidirectional_auth || session->username ||
	      session->password))
		return 0;

	if (!session->md5_tfm) {
		session->md5_tfm = crypto_alloc_tfm("md5", 0);
		if (!session->md5_tfm)
			goto fail;
	}

	/* make sure we've allocated everything we need */
	if (session->auth_client_block == NULL)
		session->auth_client_block =
			kmalloc(sizeof(*session->auth_client_block),
				GFP_KERNEL);

	if (session->auth_recv_string_block == NULL)
		session->auth_recv_string_block =
			kmalloc(sizeof(*session->auth_recv_string_block),
				GFP_KERNEL);

	if (session->auth_send_string_block == NULL)
		session->auth_send_string_block =
			kmalloc(sizeof(*session->auth_send_string_block),
				GFP_KERNEL);

	if (session->auth_recv_binary_block == NULL)
		session->auth_recv_binary_block =
			kmalloc(sizeof(*session->auth_recv_binary_block),
				GFP_KERNEL);

	if (session->auth_send_binary_block == NULL)
		session->auth_send_binary_block =
			kmalloc(sizeof(*session->auth_send_binary_block),
				GFP_KERNEL);

	/*
	 * if we have everything we need, setup the auth buffer
	 * descriptors for the login library
	 */
	session->num_auth_buffers = 0;
	memset(&session->auth_buffers, 0, sizeof(session->auth_buffers));
	if (session->auth_client_block && session->auth_recv_string_block &&
	    session->auth_send_string_block &&
	    session->auth_recv_binary_block &&
	    session->auth_send_binary_block) {
		assign_auth_buffers(session);
		return 0;
	}

 fail:
	iscsi_host_err(session, "Session requires authentication "
		       "but couldn't allocate authentication stuctures\n");
	return -ENOMEM;
}

/*
 * return value:
 * 	1: login successfully.
 *	0: Failed to login. No need to retry. Give up.
 * 	-1: Failed to login. Retry.
 */

static int
login_response_status(struct iscsi_session *session,
		      enum iscsi_login_status login_status)
{
	int ret;

	switch (login_status) {
	case LOGIN_OK:
		/* check the status class and detail */
		ret = 1;
		break;
	case LOGIN_IO_ERROR:
	case LOGIN_WRONG_PORTAL_GROUP:
	case LOGIN_REDIRECTION_FAILED:
		iscsi_disconnect(session);
		ret = -1;
		break;
	default:
		iscsi_disconnect(session);
		/*
		 * these are problems that will probably occur with any portal
		 * of this target.
		 */
		if (session->commands_queued)
			/*
			 * the session has found LUNs and been used before, so
			 * applications or the buffer cache may be expecting
			 * it to continue working.  Keep trying to login even
			 * though clearing the error may require
			 * reconfiguration on the target.
			 */
			ret = -1;
		else {
			iscsi_host_err(session, "Session giving up on login "
				       "attempts\n");
			ret = 0;
		}
	}

	return ret;
}

/*
 * return value:
 * 	2: login successfully.
 * 	1: Redirected. Retry login.
 *	0: Failed to login. No need to retry. Give up.
 * 	-1: Failed to login. Retry.
 */

static int
check_iscsi_status_class(struct iscsi_session *session, u8 status_class,
			 u8 status_detail)
{
	switch (status_class) {
	case ISCSI_STATUS_CLS_SUCCESS:
		return 2;
	case ISCSI_STATUS_CLS_REDIRECT:
		switch (status_detail) {
		case ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP:
			return 1;	/* not really success, but we want to
					 * retry immediately, with no delay
					 */
		case ISCSI_LOGIN_STATUS_TGT_MOVED_PERM:
			/*
			 * for a permanent redirect, we need to update the
			 * portal address,  and then try again.
			 */
			spin_lock(&session->portal_lock);
			/* reset the address in the current portal info */
			memcpy(&session->portal.addr, &session->addr,
			       sizeof(struct sockaddr));
			spin_unlock(&session->portal_lock);
                        return 1;       /* not really success, but we want to
                                         * retry immediately, with no delay
                                         */
		default:
			iscsi_host_err(session, "Login rejected: redirection "
				       "type 0x%x not supported\n",
				       status_detail);
			iscsi_disconnect(session);
			return -1;
		}
	case ISCSI_STATUS_CLS_INITIATOR_ERR:
		iscsi_disconnect(session);

		switch (status_detail) {
		case ISCSI_LOGIN_STATUS_AUTH_FAILED:
			iscsi_host_err(session, "Login rejected: Initiator "
				       "failed authentication with target\n");
			if ((session->num_auth_buffers < 5) &&
	    		    (session->username || session->password_length ||
			    session->bidirectional_auth))
				/*
				 * retry, and hope we can allocate the auth
				 * structures next time.
				 */
				return -1;
			else 
				return 0;	
		case ISCSI_LOGIN_STATUS_TGT_FORBIDDEN:
			iscsi_host_err(session, "Login rejected: initiator "
				       "failed authorization with target\n");
			return 0;
		case ISCSI_LOGIN_STATUS_TGT_NOT_FOUND:
			iscsi_host_err(session, "Login rejected: initiator "
				       "error - target not found (%02x/%02x)\n",
				       status_class, status_detail);
			return 0;
		case ISCSI_LOGIN_STATUS_NO_VERSION:
			/*
			 * FIXME: if we handle multiple protocol versions,
			 * before we log an error, try the other supported
			 * versions.
			 */
			iscsi_host_err(session, "Login rejected: incompatible "
				       "version (%02x/%02x), non-retryable, "
				       "giving up\n", status_class,
				       status_detail);
			return 0;
		default:
			iscsi_host_err(session, "Login rejected: initiator "
				       "error (%02x/%02x), non-retryable, "
				       "giving up\n", status_class,
				       status_detail);
			return 0;
		}
	case ISCSI_STATUS_CLS_TARGET_ERR:
		iscsi_host_err(session, "Login rejected: target error "
			       "(%02x/%02x)\n", status_class, status_detail);
		iscsi_disconnect(session);
		/*
		 * We have no idea what the problem is. But spec says initiator
		 * may retry later.
		 */
		 return -1;
	default:
		iscsi_host_err(session, "Login response with unknown status "
			       "class 0x%x, detail 0x%x\n", status_class,
			       status_detail);
		iscsi_disconnect(session);
		return 0;
	}
}

static void
initialize_iscsi_session_parameters(struct iscsi_session *session)
{
	/* logged in, get the new session ready */
	session->ever_established = 1;
	session->session_alive = 1;
	session->last_rx = jiffies;
	session->last_ping = jiffies - 1;
	session->last_window_check = jiffies;
	session->last_kill = 0;
	session->window_closed = 0;
	session->window_full = 0;
	session->nop_reply.ttt = ISCSI_RSVD_TASK_TAG;
	INIT_LIST_HEAD(&session->nop_reply_list);
	session->session_established_time = jiffies;
		/* used to detect sessions that die as soon as we hit FFP */
	session->session_drop_time = 0;
		/* used to detect sessions that aren't coming back up */
	session->login_phase_timer = 0;
}

static int
establish_session(struct iscsi_session *session)
{
	int ret = -1;
	u8 status_class;
	u8 status_detail;
	enum iscsi_login_status login_status;

	spin_lock(&session->portal_lock);
	/*
	 * Set almost everything based on the portal's settings.
	 * Don't change the address, since a temporary redirect
	 * may have already changed the address,
	 * and we want to use the redirected address rather than
	 * the portal's address.
	 */
	iscsi_set_portal_config(session);

	spin_unlock(&session->portal_lock);
	/* set a timer on the connect */
	if (session->login_timeout)
		session->login_phase_timer =
			jiffies + (session->login_timeout * HZ);

	if (iscsi_connect(session)) {
		if (signal_pending(current))
			iscsi_host_err(session, "Connect timed out\n");
		else
			iscsi_host_err(session, "Connect failed\n");

		goto done;
	}
	/*
	 * We need to grab the config_mutex before we start trying to
	 * login, to ensure update_session doesn't try to change the
	 * per-session settings while the login code is using them.  Any
	 * config updates will be deferred until after the login
	 * completes.  We grab the mutex now, so that the connect timeout
	 * will break us out if we can't get the mutex for some reason.
	 */
	if (down_interruptible(&session->config_mutex)) {
		iscsi_host_err(session, "Failed to acquire mutex before "
			       "login\n");
		goto done;
	}
	/* clear the connect timer */
	session->login_phase_timer = 0;
	if (signal_pending(current))
		flush_signals(current);
	/*
	 * try to make sure other timeouts don't go off as soon as the
	 * session is established
	 */
	session->last_rx = jiffies;
	session->last_ping = jiffies - 1;
	/* initialize session fields for the iscsi-login code */
	session->type = ISCSI_SESSION_TYPE_NORMAL;
	/*
	 * iSCSI default, unless declared otherwise by the
	 * target during login
	 */
	session->max_xmit_data_segment_len =
	    DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
	session->vendor_specific_keys = 1;
	/*
	 * use the session's rx_buffer for a login PDU buffer, since it is
	 * currently unused.  We can't afford to dynamically allocate
	 * memory right now, since it's possible we're reconnecting, and
	 * the VM system is already blocked trying to write dirty pages to
	 * the iSCSI device we're trying to reconnect.  The session's
	 * rx_buffer was sized to have enough space for us to handle the login
	 * phase.
	 */
	login_status = iscsi_login(session, session->rx_buffer,
				   sizeof(session->rx_buffer), &status_class,
				   &status_detail);
	/*
	 * release the lock on the per-session settings used by the login code
	 */
	up(&session->config_mutex);
	ret = login_response_status(session, login_status);
	if (ret < 1)
		goto done;

	/* check the login status */
	ret = check_iscsi_status_class(session, status_class, status_detail);
	if (ret < 2)
		goto done;

	initialize_iscsi_session_parameters(session);
	/* mark the session as up and accepting commands again */
	clear_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits);
	set_bit(SESSION_ESTABLISHED, &session->control_bits);
	/* wake up everyone waiting for the session to be established */
	wake_up(&session->login_wait_q);
	/* make sure we start sending commands again */
	iscsi_wake_tx_thread(TX_SCSI_COMMAND, session);
 done:
	/* clear any timer that may have been left running */
	session->login_phase_timer = 0;
	/* cleanup after a possible timeout expiration */
	if (signal_pending(current)) {
		flush_signals(current);

		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			return 0;
		else
			return -1;
	}
	return ret;
}

static char*
iscsi_strdup(char *str, int *err)
{
	int len;
	char *s;

	*err = 0;
	len = strlen(str) + 1;
	if (len == 1) {
		*err = -EINVAL;
		return NULL;
	}

	s = kmalloc(len, GFP_KERNEL);
	if (!s) {
		*err = -ENOMEM;
		return NULL;
	}

	return strcpy(s, str);
}

/*
 * return value:
 * 	1: name/alias updated. Relogin required.
 *	0: No updated needed.
 * 	-Exxx: Failed to update.
 */
static int
update_iscsi_strings(struct iscsi_session *session,
		     struct iscsi_session_ioctl *ioctld)
{
	char *iname = NULL;
	char *alias = NULL;
	char *uname = NULL;
	char *uname_in = NULL;
	char *pw = NULL;
	char *pw_in = NULL;
	int rc = 0;

	if (!ioctld->initiator_name[0]) {
		iscsi_host_err(session, "No InitiatorName\n");
		return -EINVAL;
	}
	if (strcmp(ioctld->initiator_name, session->initiator_name)) {
		iname = iscsi_strdup(ioctld->initiator_name, &rc);
		if (!iname) {
			iscsi_host_err(session, "Failed to change "
				       "InitiatorName from %s to %s\n",
				       session->initiator_name,
				       ioctld->initiator_name);
			return rc;
		}
	}

	if (ioctld->initiator_alias[0] && (!session->initiator_alias ||
	    strcmp(ioctld->initiator_alias, session->initiator_alias))) {
		alias = iscsi_strdup(ioctld->initiator_alias, &rc);
		if (!alias)
			/* Alias is not ciritical so just print an error */
			iscsi_host_err(session, "Failed to change "
					"InitiatorAlias\n");
	}

	if (ioctld->username[0] && (!session->username ||
	    strcmp(ioctld->username, session->username))) {
		uname = iscsi_strdup(ioctld->username, &rc);
		if (!uname) {
			iscsi_host_err(session, "Failed to change outgoing "
				       "username\n");
				goto failed;
		}
	}

	if (ioctld->username_in[0] && (!session->username_in ||
	    strcmp(ioctld->username_in, session->username_in))) {
		uname_in = iscsi_strdup(ioctld->username_in, &rc);
		if (!uname_in) {
			iscsi_host_err(session, "Failed to change incoming "
				       "username\n");
				goto failed;
		}
	}

	if (ioctld->password_length && (!session->password ||
	    session->password_length != ioctld->password_length ||
	    memcmp(ioctld->password, session->password,
		   session->password_length))) {
		pw = kmalloc(ioctld->password_length + 1, GFP_KERNEL);
		if (!pw) {
			iscsi_host_err(session, "Failed to change outgoing "
				       "password\n");
			rc = -ENOMEM;
			goto failed;
		}
		memcpy(pw, ioctld->password, ioctld->password_length);
	}

	if (ioctld->password_length_in && (!session->password_in ||
	    session->password_length_in != ioctld->password_length_in ||
	    memcmp(ioctld->password_in, session->password_in,
		   session->password_length_in))) {
		pw_in = kmalloc(ioctld->password_length_in + 1, GFP_KERNEL);
		if (!pw_in) {
			iscsi_host_err(session, "Failed to change incoming "
				       "password\n");
			rc = -ENOMEM;
			goto failed;
		}
		memcpy(pw_in, ioctld->password_in, ioctld->password_length_in);
	}

	if (iname) {
		kfree(session->initiator_name);	
		session->initiator_name = iname;
		rc = 1;
	}
	if (alias || (!ioctld->initiator_alias[0] &&
		      session->initiator_alias[0])) {
		kfree(session->initiator_alias);
		session->initiator_alias = alias;
		rc = 1;
	}
	if (uname || (!ioctld->username[0] && session->username)) {
		kfree(session->username);
		session->username = uname;
		rc = 1;
	}
	if (uname_in || (!ioctld->username_in[0] && session->username_in)) {
		kfree(session->username_in);
		session->username_in = uname_in;
		rc = 1;
	}
	if (pw || (!ioctld->password_length && session->password)) {
		kfree(session->password);
		session->password = pw;
		session->password_length = ioctld->password_length;
		rc = 1;
	}
	if (pw_in || (!ioctld->password_length_in && session->password_in)) {
		kfree(session->password_in);
		session->password_in = pw_in;
		session->password_length_in = ioctld->password_length_in;
		rc = 1;
	}
	return rc;
 failed:
	kfree(iname);
	kfree(alias);
	kfree(uname);
	kfree(uname_in);
	kfree(pw);
	kfree(pw_in);
	return rc;
}

int
iscsi_update_session(struct iscsi_session *session,
		     struct iscsi_session_ioctl *ioctld)
{
	int rc = 0;
	int relogin = 0;

	if (down_interruptible(&session->config_mutex)) {
		iscsi_host_err(session, "Session configuration update aborted "
			       "by signal\n");
		return -EINTR;
	}
	if (test_bit(SESSION_TERMINATED, &session->control_bits))
		return -EINVAL;

	if (ioctld->update && (ioctld->config_number < session->config_number))
		/* this update is obsolete, ignore it */
		goto err_exit;

	if (ioctld->username_in[0] || ioctld->password_length_in)
		session->bidirectional_auth = 1;
	else
		session->bidirectional_auth = 0;
	rc = allocate_auth_buffers(session);
	if (rc < 0)
		goto err_exit;

	rc = update_iscsi_strings(session, ioctld);
	if (rc > 0)
		relogin = 1;
	else if (rc < 0) 
		goto err_exit;

	session->config_number = ioctld->config_number;

	if (memcmp(session->isid, ioctld->isid, sizeof(session->isid))) {
		memcpy(session->isid, ioctld->isid, sizeof(session->isid));
		relogin = 1;
	}
	up(&session->config_mutex);

	/*
	 * the portals are guarded by a spinlock instead of the config
	 * mutex, so that we can request portal changes while a login is
	 * occuring.
	 */
	spin_lock(&session->portal_lock);
	if (iscsi_update_portal_config(session, ioctld))
		relogin = 1;
	spin_unlock(&session->portal_lock);

	if (relogin)
		if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
			spin_lock_bh(&session->task_lock);
			iscsi_request_logout(session, 3,
					     session->active_timeout);
			spin_unlock_bh(&session->task_lock);
		}
	return 0;

 err_exit:
	up(&session->config_mutex);
	return rc;
}

static int
copy_iscsi_strings(struct iscsi_session *session,
		   struct iscsi_session_ioctl *ioctld)
{
	int rc;

	session->initiator_name = iscsi_strdup(ioctld->initiator_name, &rc);
	if (rc == -EINVAL) {
		iscsi_host_err(session, "No InitiatorName\n");
		return rc;
	}
	if (rc == -ENOMEM) {
		iscsi_host_err(session, "Cannot allocate InitiatorName\n");
		return rc;
	}

	session->initiator_alias = iscsi_strdup(ioctld->initiator_alias, &rc);
	/* Alias is not ciritical so just print an error */
	if (!session->initiator_alias)
		iscsi_host_err(session, "Cannot create InitiatorAlias\n");

	session->target_name = iscsi_strdup(ioctld->target_name, &rc);
	if (rc == -EINVAL) {
		iscsi_err("No TargetName\n");
		return rc;
	}
	if (rc == -ENOMEM) {
		iscsi_host_err(session, "Cannot allocate TargetName\n");
		return rc;
	}

	session->username = iscsi_strdup(ioctld->username, &rc);
	if (rc == -ENOMEM) {
		iscsi_host_err(session, "Failed to allocate outgoing "
			       "username\n");
		return rc;
	}

	session->username_in = iscsi_strdup(ioctld->username_in, &rc);
	if (rc == -ENOMEM) {
		iscsi_host_err(session, "Failed to allocate incoming "
			       "username\n");
		return rc;
	}

	if (ioctld->password_length) {
		session->password = kmalloc(ioctld->password_length + 1,
					    GFP_KERNEL);
		if (!session->password) {
			iscsi_host_err(session, "Failed to allocate outgoing "
				       "password\n");
			return -ENOMEM;
		}
		memcpy(session->password, ioctld->password,
		       ioctld->password_length);
		session->password_length = ioctld->password_length;
	}

	if (ioctld->password_length_in) {
		session->password_in = kmalloc(ioctld->password_length_in + 1,
					       GFP_KERNEL);
		if (!session->password_in) {
			iscsi_host_err(session, "Failed to allocate incoming "
				       "password\n");
			return -ENOMEM;
		}
		memcpy(session->password_in, ioctld->password_in,
		       ioctld->password_length_in);
		session->password_length_in = ioctld->password_length_in;
	}

	return 0;
}

/**
 * reinit_session - init of session fields before attempting a re-login.
 * @session: session to initialize.
 **/
static void
reinit_session(struct iscsi_session *session)
{
	struct iscsi_nop_info *nop_info, *tmp;

	session->last_mgmt_itt = ISCSI_RSVD_TASK_TAG;
	/*
	 * can't send any nop replies now
	 */
	session->nop_reply.ttt = ISCSI_RSVD_TASK_TAG;
	list_for_each_entry_safe(nop_info, tmp, &session->nop_reply_list,
				 reply_list) {
		list_del(&nop_info->reply_list);
		kfree(nop_info);
	}

	session->logout_deadline = 0;
	session->logout_response_deadline = 0;
}

static void
init_session_structure(struct iscsi_session *session,
		       struct iscsi_session_ioctl *ioctld)
{
	INIT_LIST_HEAD(&session->list);

	/* authentication parameters */
	session->num_auth_buffers = 0;
	session->auth_client_block = NULL;
	session->auth_recv_string_block = NULL;
	session->auth_send_string_block = NULL;
	session->auth_recv_binary_block = NULL;
	session->auth_send_binary_block = NULL;

	session->socket = NULL;
	session->config_number = ioctld->config_number;
	spin_lock_init(&session->portal_lock);

	/* portal-specific parameters */
	session->portal_group_tag = -1;

	/* retry and timer specific parameters */
	sema_init(&session->config_mutex, 0);
	/* the first down should block */
	init_waitqueue_head(&session->queue_empty);
	INIT_LIST_HEAD(&session->pending_queue);
	INIT_LIST_HEAD(&session->active_queue);
	INIT_LIST_HEAD(&session->done_queue);
	spin_lock_init(&session->task_lock);
	INIT_LIST_HEAD(&(session->tx_task_head));
	atomic_set(&session->num_active_tasks, 0);
	init_waitqueue_head(&session->tx_wait_q);
	init_waitqueue_head(&session->login_wait_q);
	sema_init(&session->tx_blocked, 0);

	session->next_itt = 1;
	session->ever_established = 0;
	session->session_alive = 0;
	session->time2wait = -1;
	session->logout_itt = ISCSI_RSVD_TASK_TAG;
	session->last_mgmt_itt = ISCSI_RSVD_TASK_TAG;
	session->mgmt_task_complete = NULL;
	session->session_drop_time = 0;
}

static int
check_login_timeouts(struct iscsi_session *session)
{
	unsigned long session_timeout = 0;

	if (session->login_phase_timer) {
		session_timeout = session->login_phase_timer;
		if (time_before_eq(session_timeout, jiffies)) {
			iscsi_host_err(session, "Login phase timed out, timeout"
				       " was set for %lu\n", session_timeout);
			session->login_phase_timer = 0;
			session_timeout = 0;
			signal_iscsi_threads(session);
		}
	}
	return session_timeout;
}

static unsigned long
check_session_ping_timeout(struct iscsi_session *session,
			   unsigned long timeout)
{
	/* should have received something by now, kill the connection */
	if ((session->last_kill == 0) ||
	    time_before_eq(session->last_kill + HZ, jiffies)) {
		session->last_kill = jiffies;
		iscsi_host_err(session, "%lu second timeout expired rx %lu, "
			       "ping %lu, now %lu\n",
			       timeout + session->ping_timeout,
			       session->last_rx, session->last_ping, jiffies);
		iscsi_drop_session(session);
		return (jiffies + HZ);
	} else
		return 0;
}

static unsigned long
check_session_idle_timeout(struct iscsi_session *session)
{
	unsigned long active_timeout, idle_timeout;

	/*
	 * session is idle, but may become active without the timer being
	 * notified, so use smaller of (now + active_timeout, last_rx +
	 * idle_timeout)
	 */
	idle_timeout = session->last_rx + (session->idle_timeout * HZ);
	active_timeout = jiffies + (session->active_timeout * HZ);
	if (time_before_eq(idle_timeout, active_timeout))
		return idle_timeout;
	else
		return active_timeout;

}

static unsigned long
handle_iscsi_logout_timeout(struct iscsi_session *session)
{
	unsigned long session_timeout = 0;

	/*
	 * we're waiting for tasks to complete before logging out. No need to
	 * check the CmdSN window, since we won't be starting any more tasks.
	 */
	if (time_before_eq(session->logout_response_deadline, jiffies)) {
		/*
		 * passed the deadline for a logout response, just drop the
		 * session
		 */
		iscsi_host_err(session, "Logout response timeout, dropping "
			       "session\n");
		session->logout_response_deadline = 0;
		session->logout_deadline = 0;
		iscsi_drop_session(session);
	} else {
		if (time_before_eq(session->logout_deadline, jiffies)) {
			/* send a logout */
			session->logout_deadline = 0;
			iscsi_wake_tx_thread(TX_LOGOUT, session);
			if ((session_timeout == 0) ||
			    time_before(session->logout_response_deadline,
					session_timeout))
				session_timeout = session->logout_response_deadline;
		} else {
			if ((session_timeout == 0) ||
			    time_before(session->logout_deadline,
					session_timeout))
				session_timeout = session->logout_deadline;
		}
	}
	return session_timeout;
}

static void
check_iscsi_cmd_window_closed(struct iscsi_session *session,
			      unsigned long *session_timeout)
{
	if (test_bit(SESSION_WINDOW_CLOSED, &session->control_bits) &&
	    time_before_eq(session->last_window_check + HZ, jiffies)) {
		/*
		 * command window closed, ping once a second to ensure
		 * we find out when it re-opens.  Target should send
		 * us an update when it does, but we're not very
		 * trusting of target correctness.
		 */
		session->last_window_check = jiffies;
		/* request a window update from the target with Nops */
		iscsi_wake_tx_thread(TX_PING, session);

		if (((*session_timeout) == 0) ||
		    time_before(session->last_window_check + HZ,
				*session_timeout))
			*session_timeout = session->last_window_check + HZ;
	}
}

static int
check_full_feature_timeouts(struct iscsi_session *session)
{
	unsigned long timeout;
	unsigned long session_timeout = 0;

	if (atomic_read(&session->num_active_tasks))
		timeout = session->active_timeout;
	else
		timeout = session->idle_timeout;

	if (timeout) {
		if (session->ping_timeout &&
		    time_before_eq(session->last_rx + (timeout * HZ) +
				   (session->ping_timeout * HZ), jiffies))
			session_timeout = check_session_ping_timeout
						(session, timeout);
		else {
			if (time_before_eq
			    (session->last_rx + (timeout * HZ), jiffies)) {
				if (time_before_eq
				    (session->last_ping, session->last_rx)) {
					/*
					 * send a ping to try to provoke some
					 * traffic
					 */
					session->last_ping = jiffies - 1;
					iscsi_wake_tx_thread(TX_PING, session);
				}
				session_timeout =
					session->last_rx + (timeout * HZ) +
					(session->ping_timeout * HZ);
			} else {
				if (atomic_read(&session->num_active_tasks))
					session_timeout =
						session->last_rx +
						(session->active_timeout * HZ);
				else
					session_timeout =
						check_session_idle_timeout
						(session);
			}
		}
	}
	if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits))
		session_timeout = handle_iscsi_logout_timeout(session);
	else
		check_iscsi_cmd_window_closed(session, &session_timeout);

	return session_timeout;
}

/*
 * do timer processing for one session, and return the length of time
 * (in jiffies) til this session needs to be checked again.
 */
static void
session_timeout(unsigned long data)
{
	struct iscsi_session *session = (struct iscsi_session *)data;
	unsigned long timeout = 0;

	if (test_bit(SESSION_TERMINATED, &session->control_bits))
		return;

	if (!test_bit(SESSION_ESTABLISHED, &session->control_bits))
		/* check login phase timeouts */
		timeout = check_login_timeouts(session);
	else
		/* check full-feature phase timeouts. */
		timeout = check_full_feature_timeouts(session);

	if (!timeout)
		timeout = DEF_SESSION_TIMEOUT;
	mod_timer(&session->session_timer, timeout);
}

void
iscsi_wake_tx_thread(int control_bit, struct iscsi_session *session)
{
	set_bit(control_bit, &session->control_bits);
	set_bit(TX_WAKE, &session->control_bits);
	wake_up(&session->tx_wait_q);
}

/**
 * iscsi_wait_for_session - Wait for a session event to be established.
 * @session: session to wait on.
 * @ignore_timeout: If zero this will return when the replacement timeout fires.
 *
 * Description:
 *    Returns 1 to indicate sesssion was established, or 0 to indicate
 *    we timed out (if ignore_timeout == 0) or are terminating.
 **/
int
iscsi_wait_for_session(struct iscsi_session *session, int ignore_timeout)
{
	int rc = 0;

	while (1) {
		wait_event_interruptible(session->login_wait_q,
			test_bit(SESSION_ESTABLISHED, &session->control_bits) ||
			test_bit(SESSION_TERMINATING, &session->control_bits) ||
			(!ignore_timeout &&
			 test_bit(SESSION_REPLACEMENT_TIMEDOUT,
				  &session->control_bits)));

		if (signal_pending(current))
			flush_signals(current);

		/*
		 * need to test for termnination first to avoid falling
		 * in the tx request loop for ever
		 */
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			break;

		if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
			rc = 1;
			break;
		}

		if (!ignore_timeout && test_bit(SESSION_REPLACEMENT_TIMEDOUT,
						 &session->control_bits))
			break;
	}

	return rc;
}

/*
 * TODO: fix properly (kill this crap) when more than one R2T
 * or MCS is supported
 */
static struct {
	void (* request_fn)(struct iscsi_session *);
} tx_request_fns[] = {
	{ iscsi_send_nop_out },
	{ iscsi_send_task_mgmt },
	{ iscsi_run_pending_queue },
	{ iscsi_send_nop_replys },
	{ iscsi_send_r2t_data },
	{ iscsi_send_logout },
};

static void
wait_for_tx_requests(struct iscsi_session *session)
{
	int req;

	wait_event_interruptible(session->tx_wait_q,
		test_and_clear_bit(TX_WAKE, &session->control_bits));

	for (req = 0; req < TX_WAKE; req++) {
		if (signal_pending(current))
			return;
		/*
		 * when a logout is in progress or about to be sent
		 * we do not start new requests, but we continue to
		 * respond to R2Ts and Nops.
		 */
		if (test_and_clear_bit(req, &session->control_bits)) {
			if (test_bit(SESSION_LOGOUT_REQUESTED,
				     &session->control_bits) &&
			    req <= TX_SCSI_COMMAND)
				continue;

			tx_request_fns[req].request_fn(session);
		}
	}
}

/**
 * session_kthread_sleep - put a thread to sleep while waiting for shutdown.
 * @session: session. 
 *
 * Description:
 *    If for some reason we could not relogin into a session we sleep here
 *    and and wait for someone to remove the session. Returns -EPERM to
 *    indicate the thread should exit, or zero to indicate that the thread
 *    can proceed with its normal action.
 **/
static inline int
session_kthread_sleep(struct iscsi_session *session)
{
 retest:
	set_current_state(TASK_INTERRUPTIBLE);
	if (kthread_should_stop()) {
		__set_current_state(TASK_RUNNING);
		return -EPERM;
	}

	/*
	 * We fall into this sleep, when someone has broken us
	 * out of the lower loops that process requests or log us in,
	 * terminate the session (session drops will not sleep here),
	 * but have not (yet) cleaned up the host and called kthread_stop()).
	 */
	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		schedule();
		if (signal_pending(current))
			flush_signals(current);
		goto retest;
	}
	__set_current_state(TASK_RUNNING);
	return 0;
}

/*
 * move to session timeout handler when that is fixed??
 */
static void
replacement_timed_out(unsigned long data)
{
	struct iscsi_session *session = (struct iscsi_session *)data;

	iscsi_host_err(session, "replacement session time out after %d "
		       "seconds, drop %lu, now %lu, failing all commands\n",
		       session->replacement_timeout,
		       session->session_drop_time, jiffies);

	/*
	 * need a lock or some real state code.
	 */
	if (test_bit(SESSION_ESTABLISHED, &session->control_bits) ||
	    test_and_set_bit(SESSION_REPLACEMENT_TIMEDOUT,
			     &session->control_bits))
		return;

	/*
	 * No point in letting the commands timeout, fail them.
	 */
	spin_lock(&session->task_lock);
	iscsi_flush_queues(session, ISCSI_MAX_LUNS, DID_BUS_BUSY);
	spin_unlock(&session->task_lock);
	wake_up_all(&session->login_wait_q);
}

/* the writer thread */
static int
iscsi_tx_thread(void *data)
{
	int rc;
	unsigned long tmo;
	struct iscsi_session *session = (struct iscsi_session *)data;
	struct timer_list replacement_timer;

	init_timer(&replacement_timer);
	replacement_timer.data = (unsigned long)session;
	replacement_timer.function = replacement_timed_out;
	
	current->flags |= PF_MEMALLOC;
	allow_signal(SIGHUP);

	/*
	 * tell the rx thread that we're about to block, and that
	 * it can safely call iscsi_sendmsg now as part of
	 * the Login phase.
	 */
	up(&session->tx_blocked);

	while (!session_kthread_sleep(session)) {
		tmo = session->replacement_timeout * HZ;

		if (tmo && session->session_drop_time) {
			replacement_timer.expires = jiffies + tmo;
			add_timer(&replacement_timer);
		}

		rc = iscsi_wait_for_session(session, 1);
		del_timer_sync(&replacement_timer);
		if (!rc)
			continue;

		/* we're up and running with a new session */
		down(&session->tx_blocked);

		/* make sure we start sending commands again */
		set_bit(TX_PING, &session->control_bits);
		set_bit(TX_SCSI_COMMAND, &session->control_bits);
		set_bit(TX_WAKE, &session->control_bits);

		while (!signal_pending(current))
			wait_for_tx_requests(session);
		flush_signals(current);

		up(&session->tx_blocked);
	}

	return 0;
}

static int
rx_establish_session(struct iscsi_session *session, unsigned int login_delay)
{
	int rc;
	unsigned long login_failures = 0;

	while (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		if (login_delay) {
			iscsi_host_notice(session, "Waiting %u seconds before "
					  "next login attempt\n", login_delay);
			msleep_interruptible(login_delay * 1000);
		}

		/*
		 * wait for the tx thread to block or exit, ignoring signals.
		 * the rx thread needs to know that the tx thread is not
		 * running before it can safely close the socket and start
		 * a new login phase on a new socket, Also, tasks still in
		 * use by the tx thread can't safely be completed on
		 * a session drop.
		 */
		down(&session->tx_blocked);
		up(&session->tx_blocked);

		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			return 0;

		iscsi_disconnect(session);

		clear_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
		clear_bit(SESSION_WINDOW_CLOSED, &session->control_bits);

		rc = establish_session(session);
		if (rc > 0) {
			/* established or redirected */
			login_failures = 0;
			session->session_alive = 1;
		} else if (rc < 0)
			/* failed, retry */
			login_failures++;
		 else {
			/* failed, give up */
			iscsi_host_err(session, "Session giving up\n");
			set_bit(SESSION_TERMINATING, &session->control_bits);
			return 0;
		}

		/* slowly back off the frequency of login attempts */
		if (login_failures == 0)
			login_delay = 0;
		else if (login_failures < 30)
			login_delay = 1;
		else if (login_failures < 48)
			login_delay = 5;
		else if (session->replacement_timeout &&
			 time_before_eq(session->session_drop_time +
					(HZ * session->replacement_timeout),
					jiffies))
			login_delay = 10;
		 else
			login_delay = 60;
	}

	return 1;
}

/**
 * get_time2wait - return iSCSI DefaultTime2Wait
 * @session: iscsi session
 * @short_sessions: number of short sessions
 *
 * Description:
 *   Return DefaultTime2Wait. However, if the session dies really
 *   quicky after we reach FFP, we'll not be interoperable due to bugs
 *   in the target (or this driver) that send illegal opcodes,
 *   or disagreements about how to do CRC  calculations. To
 *   avoid spinning, we track sessions with really short
 *   lifetimes, and decrease the login frequency if we keep
 *   getting session failures, like we do for login failures.
 **/
static unsigned int
get_time2wait(struct iscsi_session *session, unsigned long *short_sessions)
{
	unsigned int login_delay = 0;

	if (session->time2wait >= 0) {
		login_delay = session->time2wait;
		session->time2wait = -1;
	} else
		login_delay = session->def_time2wait;

	if (time_before_eq(session->session_drop_time,
		     	   session->session_established_time + (2 * HZ))) {
		(*short_sessions)++;

		if (*short_sessions < 30)
			login_delay = max_t(unsigned int, login_delay, 1);
		else if (*short_sessions < 48)
			login_delay = max_t(unsigned int, login_delay, 5);
		else if (session->replacement_timeout &&
			 time_before_eq(session->session_drop_time +
					(HZ * session->replacement_timeout),
					jiffies))
			login_delay = max_t(unsigned int, login_delay, 10);
		 else
			login_delay = max_t(unsigned int, login_delay, 60);

		iscsi_host_warn(session, "Session has ended quickly %lu times, "
				"login delay %u seconds\n", *short_sessions,
				login_delay);
	} else
		/* session lived long enough that the target is probably ok */
		*short_sessions = 0;

	return login_delay;
}

static int
iscsi_rx_thread(void *data)
{
	struct iscsi_hdr hdr;
	unsigned int login_delay = 0;
	unsigned long short_sessions = 0;
	struct iscsi_session *session = (struct iscsi_session *)data;
	current->flags |= PF_MEMALLOC;
	allow_signal(SIGHUP);

        while (!session_kthread_sleep(session)) {
		if (!rx_establish_session(session, login_delay))
			continue;

		while (!signal_pending(current))
			iscsi_recv_pdu(session, &hdr, session->header_digest,
				       session->rx_buffer, ISCSI_RXCTRL_SIZE,
				       session->data_digest, 0);
		flush_signals(current);

		login_delay = get_time2wait(session, &short_sessions);
		/*
		 * we need to wait for the tx thread to block before
		 * trying to complete commands, since it may be using a
		 * task at the moment, which means we can't complete it yet.
		 * even if the session is terminating, we must wait for
		 * the tx thread.
		 */
		down(&session->tx_blocked);
		up(&session->tx_blocked);

		spin_lock_bh(&session->task_lock);
		/*
		 * session dropped unexpectedly, often due to
		 * network problems
		 */
		iscsi_host_err(session, "Session dropped\n");
		iscsi_flush_queues(session, ISCSI_MAX_LUNS, DID_BUS_BUSY);
		reinit_session(session);
		spin_unlock_bh(&session->task_lock);
	}
	/*
	 * If there are any commands left this will remove them.
	 */
	spin_lock_bh(&session->task_lock);
	iscsi_flush_queues(session, ISCSI_MAX_LUNS, DID_NO_CONNECT);
	spin_unlock_bh(&session->task_lock);

	return 0;
}

static int
start_session_threads(struct iscsi_session *session)
{
	session->tx_task = kthread_run(iscsi_tx_thread, session, "iscsi-tx");
	if (IS_ERR(session->tx_task)) {
		iscsi_host_err(session, "Failed to start tx thread, terminating"
			       " session\n");
		goto fail;
	}

	session->rx_task = kthread_run(iscsi_rx_thread, session, "iscsi-rx");
	if (IS_ERR(session->rx_task)) {
		iscsi_host_err(session, "Failed to start rx thread, terminating"
			       " session\n");
		goto shutdown_tx_thread;
	}

	init_timer(&session->session_timer);
	session->session_timer.data = (unsigned long)session;
	session->session_timer.function = session_timeout;
	/*
	 * During periods of transition we use some safe value.
	 */
	session->session_timer.expires = DEF_SESSION_TIMEOUT;
	add_timer(&session->session_timer);

	return 0;

 shutdown_tx_thread:
	set_bit(SESSION_TERMINATING, &session->control_bits);
	kthread_stop(session->tx_task);
 fail:
	return -EAGAIN;
}

void
iscsi_destroy_session(struct iscsi_session *session)
{
	set_bit(SESSION_TERMINATING, &session->control_bits);
	down(&iscsi_session_sem);
	list_del(&session->list);
	up(&iscsi_session_sem);
	clear_bit(SESSION_ESTABLISHED, &session->control_bits);

	session->session_drop_time = jiffies ? jiffies : 1;
	signal_iscsi_threads(session);

	kthread_stop(session->tx_task);
	kthread_stop(session->rx_task);

	if (session->socket)	
		iscsi_disconnect(session);

	set_bit(SESSION_TERMINATED, &session->control_bits);
	/*
	 * wake up any ioctls sleeping on the session
	 * (this will be needed when update_session is fixed)
	 */
	wake_up(&session->login_wait_q);
	del_timer_sync(&session->session_timer);

	/*
	 * grab the config mutex to make sure update_session is not
	 * accessing the session fields we are going to free
	 */
	down(&session->config_mutex);
	free_session(session);
	up(&session->config_mutex);
}

int
iscsi_create_session(struct iscsi_session *session,
		     struct iscsi_session_ioctl *ioctld)
{
	int rc;

	init_session_structure(session, ioctld);

	session->preallocated_task = kmem_cache_alloc(iscsi_task_cache,
						      GFP_KERNEL);
	if (!session->preallocated_task) {
		iscsi_host_err(session, "Couldn't preallocate task\n");
		rc = -ENOMEM;
		goto free_session;
	}

	session->mgmt_task = kmem_cache_alloc(iscsi_task_cache, GFP_KERNEL);
	if (!session->mgmt_task) {
		iscsi_host_err(session, "Couldn't preallocate mgmt task\n");
		rc = -ENOMEM;
		goto free_session;
	}
	memset(session->mgmt_task, 0, sizeof(*session->mgmt_task));
	iscsi_init_task(session->mgmt_task);

	rc = copy_iscsi_strings(session, ioctld);
	if (rc)
		goto free_session;

	memcpy(session->isid, ioctld->isid, sizeof(session->isid));

	/*
	 * FIXME: Do we have to check on both the username_in and
	 * password_length_in. Same with iscsi_update_session as well? Smitha
	 */
	if (ioctld->username_in[0] || ioctld->password_length_in)
		session->bidirectional_auth = 1;
	else
		session->bidirectional_auth = 0;
	rc = allocate_auth_buffers(session);
	if (rc)
		goto free_session;

	memcpy(&session->portal, &ioctld->portal, sizeof(ioctld->portal));
	iscsi_set_portal(session);

	/*
	 * preallocate rx/tx_tfm, so that we do not have to possibly
	 * call crypto_alloc_tfm (it uses GFP_KERNEL) while IO is queued.
	 */
	session->rx_tfm = crypto_alloc_tfm("crc32c", 0);
	if (!session->rx_tfm) {
		rc = -ENOMEM;
		goto free_session;
	}

	session->tx_tfm = crypto_alloc_tfm("crc32c", 0);
	if (!session->tx_tfm) {
		rc = -ENOMEM;
		goto free_session;
	}

	rc = start_session_threads(session);
	up(&session->config_mutex);
	if (rc)
		goto free_session;

	down(&iscsi_session_sem);
	list_add_tail(&session->list, &iscsi_sessions);
	up(&iscsi_session_sem);

	wait_event_interruptible(session->login_wait_q,
		test_bit(SESSION_ESTABLISHED, &session->control_bits));
	if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		iscsi_destroy_session(session);
		return -ENOTCONN;
	}

	return 0;
	
 free_session:
	free_session(session);
	return rc;
}

struct iscsi_session *
iscsi_find_session(const char *target_name, u8 isid[6], int tpgt)
{
	struct iscsi_session *session;

	down(&iscsi_session_sem);

	list_for_each_entry(session, &iscsi_sessions, list) {
		 if (!strcmp(session->target_name, target_name) &&
		     !memcmp(session->isid, isid, sizeof(isid)) &&
		     session->portal_group_tag == tpgt) {
			if (scsi_host_get(session->shost)) {
				up(&iscsi_session_sem);
				return session;
			}
			break;
		}
	}

	up(&iscsi_session_sem);
	return NULL;
}

int
iscsi_update_address(struct iscsi_session *session, char *address)
{
	struct sockaddr_in *addr;
	char *tag;
	char *port;
	int err = 1;

	tag = strrchr(address, ',');
	if (tag) {
		*tag = '\0';
		tag++;
	}

	port = strrchr(address, ':');
	if (port) {
		*port = '\0';
		port++;
	}

	/*
	 * Still only ipv4 is supported. No access to ipv6
	 * to test so feel free to implement it later...
	 */
	if (address[0] == '[') {
		err = 0;
		goto done;
	}

	addr = (struct sockaddr_in *)&session->addr;
	addr->sin_addr.s_addr = in_aton(address);
	if (port)
		addr->sin_port = htons(simple_strtoul(port, NULL, 0));
	else
		addr->sin_port = htons(ISCSI_LISTEN_PORT);

 done:
	/* restore the original strings */
	if (tag) {
		--tag;
		*tag = ',';
	}

	if (port) {
		--port;
		*port = ':';
	}

	return err;
}
