/*
 * include/linux/iosem.h
 *
 * Copyright (C) 2005 Trond Myklebust <Trond.Myklebust@netapp.com>
 *
 * Definitions for iosems. These can act as mutexes, but unlike
 * semaphores, their code is 100% arch-independent, and can therefore
 * easily be expanded in order to provide for things like
 * asynchronous I/O.
 */

#ifndef __LINUX_SEM_LOCK_H
#define __LINUX_SEM_LOCK_H

#ifdef __KERNEL__
#include <linux/wait.h>
#include <linux/workqueue.h>

/*
 * struct iosem: iosem mutex
 *      state: bitmask - currently only signals whether or not an exclusive
 *      		 lock has been taken
 *      wait: FIFO wait queue
 */
struct iosem {
#define IOSEM_LOCK_EXCLUSIVE (31)
/* #define IOSEM_LOCK_SHARED (30) */
	unsigned long state;
	wait_queue_head_t wait;
};



/*
 * struct iosem_wait: acts as a request for a lock on the iosem
 *      lock: backpointer to the iosem
 *      wait: wait queue entry. note that the callback function
 *            defines what to do when the lock has been granted
 */
struct iosem_wait {
	struct iosem *lock;
	wait_queue_t wait;
};

/*
 * struct iosem_work: used by asynchronous waiters.
 *
 * 	work: work to schedule once the iosem has been granted. The
 * 	      function containing the critical code that needs to
 * 	      run under the protection of the lock should be placed here.
 * 	      The same function is responsible for calling iosem_unlock()
 * 	      when done.
 * 	waiter: iosem waitqueue entry
 */
struct iosem_work {
	struct work_struct work;
	struct iosem_wait waiter;
};

/*
 * Functions for synchronous i/o
 */

/* Synchronously grab an iosem.
 * These functions act in pretty much the same way down()/up()
 * do for semaphores.
 */
extern void FASTCALL(iosem_lock(struct iosem *lk));
extern void FASTCALL(iosem_unlock(struct iosem *lk));

/*
 * Callback function to wake up the sleeping task once
 * it has been granted an exclusive lock
 */
extern int iosem_lock_wake_function(wait_queue_t *wait, unsigned mode, int sync, void *key);

/* Initialize a struct iosem in the "unlocked" state */
static inline void iosem_init(struct iosem *lk)
{
	lk->state = 0;
	init_waitqueue_head(&lk->wait);
}

/* Initializes a lock request */
static inline void iosem_waiter_init(struct iosem_wait *waiter)
{
	waiter->lock = NULL;
	init_waitqueue_entry(&waiter->wait, current);
	INIT_LIST_HEAD(&waiter->wait.task_list);
}

/* 
 * Functions for asynchronous I/O.
 */

/* Requests an exclusive lock on the iosem on behalf of a workqueue entry "wk".
 * Schedule wk->work for execution as soon as the lock is granted. */
extern int FASTCALL(iosem_lock_and_schedule_work(struct iosem *lk, struct iosem_work *wk));

/* Waitqueue notifier that schedules work once the exclusive lock has
 * been granted */
extern int iosem_lock_and_schedule_function(wait_queue_t *wait, unsigned mode, int sync, void *key);

static inline void iosem_work_init(struct iosem_work *wk, void (*func)(void *), void *data)
{
	INIT_WORK(&wk->work, func, data);
}

#endif /* __KERNEL__ */
#endif /* __LINUX_SEM_LOCK_H */
